# B ch. level adjustment tool for Yellowed Photo (Linear) Ver. 0.1
# (c)2023 Ohnishi, Yasuo 
# 2023.11
from ij import IJ, ImagePlus
from ij import plugin

from ij import WindowManager as WM
from ij.io import OpenDialog
from ij.plugin import ChannelSplitter
from ij.plugin import Duplicator
from ij.process import ImageConverter, ImageProcessor, FloatProcessor
from ij.gui import GenericDialog
from ij.plugin.filter import ThresholdToSelection
from loci.plugins import BF
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')


#-----------------
def set_Lmax(imp):
    # set Maximum value of Lightness
    BitDepth = imp.getBitDepth()
    if BitDepth == 32:
        L_max = imp.getDisplayRangeMax()
        if int(L_max) > 65536:
            L_max = 16777215
        elif int(L_max) > 256:
            L_max = 65535
        elif int(L_max) > 1.01:
            L_max = 255
        else:
            L_max = 1.0
    elif BitDepth == 16:
        L_max = 65535
    else: # 8 or 24 bit
        L_max = 255
    return L_max


#--- image type detection
def imageTypeDetection(imp):
    StackSize = imp.getStackSize()
    BitDepth = imp.getBitDepth()
    return StackSize, BitDepth

#----
def sRGB_TRC_encode(imp, L_max):
    impTrc1 = Duplicator().run(imp)
    impTrc2 = Duplicator().run(imp)
    cshadowThld = 0.0031308 * L_max
    # non shadow region
    impmask = imageBinarySmpl(impTrc1, cshadowThld, L_max, L_max, \
                          "non_shadow_mask")
    impTrc1.setDisplayRange(0.0, L_max)
    impTrc1.getProcessor().gamma(1.0 / 2.4)
    impTrc1.getProcessor().multiply(1.055)
    impTrc1.getProcessor().subtract(0.055 * L_max)
    ic = plugin.ImageCalculator()
    imp_nonShadow =  ic.run("Min create", impmask, impTrc1)

    # shadow region
    impmaskShadow = imageBinarySmpl(impTrc2, 0.0, cshadowThld, L_max, \
                                "shadow_mask")
    impTrc2.setDisplayRange(0.0, L_max)
    impTrc2.getProcessor().multiply(12.92)
    imp_Shadow =  ic.run("Min create", impmaskShadow, impTrc2)

    # shadow + non shadow
    impTrc3 =  ic.run("Max create", imp_Shadow, imp_nonShadow)
    impTrc3.setTitle("sRGB TRC encoded")
    impmask.close()
    impmaskShadow.close()
    imp_Shadow.close()
    imp_nonShadow.close()
    impTrc1.close()
    impTrc2.close()
    return impTrc3

#--- TRC en / decode -------

def sRGB_TRC_encode(imp, L_max):
    impTrc1 = Duplicator().run(imp)
    impTrc2 = Duplicator().run(imp)
    cshadowThld = 0.0031308 * L_max
    # non shadow region
    impmask = imageBinarySmpl(impTrc1, cshadowThld, L_max, L_max, \
                          "non_shadow_mask")
    impTrc1.setDisplayRange(0.0, L_max)
    impTrc1.getProcessor().gamma(1.0 / 2.4)
    impTrc1.getProcessor().multiply(1.055)
    impTrc1.getProcessor().subtract(0.055 * L_max)
    ic = plugin.ImageCalculator()
    imp_nonShadow =  ic.run("Min create", impmask, impTrc1)

    # shadow region
    impmaskShadow = imageBinarySmpl(impTrc2, 0.0, cshadowThld, L_max, \
                                "shadow_mask")
    impTrc2.setDisplayRange(0.0, L_max)
    impTrc2.getProcessor().multiply(12.92)
    imp_Shadow =  ic.run("Min create", impmaskShadow, impTrc2)

    # shadow + non shadow
    impTrc3 =  ic.run("Max create", imp_Shadow, imp_nonShadow)
    impTrc3.setTitle("sRGB TRC encoded")
    impmask.close()
    impmaskShadow.close()
    imp_Shadow.close()
    imp_nonShadow.close()
    impTrc1.close()
    impTrc2.close()
    return impTrc3

#-----------

def sRGB_TRC_decode(imp, L_max):
    impTrc1 = Duplicator().run(imp)
    impTrc2 = Duplicator().run(imp)
    cshadowThld = 0.04045 * L_max
    BitDepth = imp.getBitDepth()
    ImageConverter(imp).convertToGray32()
    # not shadow region
    impmask = imageBinarySmpl(impTrc1, cshadowThld, L_max, L_max, \
                          "shadow_mask")
        # ROI: white back: black
    impTrc1.setDisplayRange(0.0, L_max)
    impTrc1.getProcessor().add(0.055 * L_max)
    impTrc1.getProcessor().multiply(1 / 1.055)
    impTrc1.getProcessor().gamma(2.4)
    ic = plugin.ImageCalculator()
    imp_nonShadow =  ic.run("Min create", impmask, impTrc1)
    # shadow region
    impmaskShadow = imageBinarySmpl(impTrc2, 0.0, cshadowThld, L_max, \
                                "non_shadow_mask")
    impTrc2.setDisplayRange(0.0, L_max)
    impTrc2.getProcessor().multiply(1 / 12.92)
    imp_Shadow =  ic.run("Min create", impmaskShadow, impTrc2)
    impTrc3 =  ic.run("Max create", imp_Shadow, imp_nonShadow)
    if BitDepth == 16:
        ImageConverter(impTrc3).convertToGray16()
        ImageConverter(imp).convertToGray16()

    elif BitDepth == 8 or BitDepth == 24:
        ImageConverter(impTrc3).convertToGray8()
        ImageConverter(imp).convertToGray8()
    impmask.close()
    impmaskShadow.close()
    imp_Shadow.close()
    imp_nonShadow.close()
    impTrc1.close()
    impTrc2.close()

    return impTrc3


#-----------------
def imageBinarySmpl(impOrg, lowerV, upperV, roiColor, maskname):
    # fill roiColor between lowerV - upperV
    # make binary mask image with setThreshold
    # lowerV: lowerThreshold / upperV: upperThreshold (double)
    # roiColor: Color Value, ROI filled with (double)
    # maskname: name of mask image (string)
    imp = Duplicator().run(impOrg)
    ip = imp.getProcessor()
    mask = ip.createProcessor(ip.width, ip.height) # Background Value = 0.0
    mask.setValue(roiColor) 
                    # with which color ROI filled
    ip.setThreshold(lowerV, \
        upperV, ImageProcessor.NO_LUT_UPDATE)
    roi = ThresholdToSelection.run(imp)
    mask.setRoi(roi)
    mask.fill(roi)
    maskimp = ImagePlus(maskname, mask)
    imp.close()
    stat = maskimp.getStatistics()
    return maskimp

#----load image-------
def load_image():
    imgCurrent = WM.getCurrentImage()
    if not imgCurrent:
        od = OpenDialog("Choose a file")
        folder = od.getDirectory()
        filename = od.getFileName()
    # Always Open file with Bio-format Plugin
        if filename:
            imps = BF.openImagePlus(folder + filename) 
            imp0 = imps[0] #imp0: Original Image 
            imp0.show()
        else:
            IJ.showMessage("File is not selected.")
    else: # target image is already open
        imp0 = imgCurrent
        filename = imp0.getTitle()
        try:
            folder = imp0.getOriginalFileInfo().directory
        except:
            folder = IJ.getDirectory("Output_directory")
    return imp0, filename, folder
#----
#-----------------------------
def get_Choice():
                # dialog of selecting target channel for mask valid lightness range
    gd = GenericDialog("sRGB TRC en/decoder")
    gd.addMessage("encode: Linear to sRGB TRC")
    gd.addMessage("decode: sRGB TRC to Linear")
    processChoice = ["encode", "decode"]
    gd.addChoice("Select applying process", processChoice, processChoice[1])
  
    #---- show the dialog ----
    gd.showDialog()

    process =  gd.getNextChoice()
    if gd.wasCanceled():
        process = "cancel"

    return process
#-----------
def get_cropped_imp_stat(imp):
    width = imp.getWidth()
    height = imp.getHeight()
    roi_x = int(width * 0.15)
    roi_y = int(height * 0.15)
    roi_width = int(width * 0.7)
    roi_height = int(height * 0.7)

    imp.setRoi(roi_x, roi_y, roi_width, roi_height)
    impTmp = imp.crop()
    stat = impTmp.getStatistics()
    impTmp.close()
    imp.resetRoi()
    return stat

#---  code for main ---
try:
    imp0, filename, folder = load_image()
    if filename:
        StackSize, BitDepth = imageTypeDetection(imp0)
except:
    BitDepth = 8
    StackSize = 0
    IJ.showMessage("User canceled.")
if BitDepth == 8 or BitDepth == 24:
    if StackSize != 0:
        IJ.showMessage\
            ("Input image must be 16 or 32bit. Job will be canceled.")
    process = "cancel"
else:
    process = "go"
if process != "cancel":
    if filename:
        IJ.log("folder: " + folder)
        IJ.log("filename: " + filename)
        IJ.log("StackSize: " + str(StackSize))
        IJ.log("BitDepth: " + str(BitDepth))
    n = imp0.getStackSize()
    L_max = set_Lmax(imp0)
    imp1 = Duplicator().run(imp0)
    impsCh1 = ChannelSplitter.split(imp1)
    impsCh2 = []
    for i in range(n):
        i = sRGB_TRC_decode(impsCh1[i], L_max)
        impsCh2.append(i)
    if BitDepth == 16:
        for i in range(n):
            ImageConverter(impsCh2[i]).convertToGray16()
    elif BitDepth == 32:
        for i in range(n):
            ImageConverter(impsCh2[i]).convertToGray32()
    chName = ['R', 'G', 'B'] # Channel name
    statR = get_cropped_imp_stat(impsCh2[0])
    statG = get_cropped_imp_stat(impsCh2[1])
    statB = get_cropped_imp_stat(impsCh2[2])
    RGmean = (statR.mean + statG.mean) / 2.0
    if RGmean * 0.8  > statB.mean:
        factor = RGmean * 0.8 / statB.mean
        impsCh2[2].getProcessor().multiply(factor)
        impsCh2[2].getProcessor().gamma(0.75)
        statB2 =  get_cropped_imp_stat(impsCh2[2])
        if statG.max > statB2.max:
            factor = 1.05 * statG.max / statB2.max
            impsCh2[2].getProcessor().multiply(factor)
        print (str(statB2.histogram))

    impsCh3 = []
    for i in range(2):
        impsCh3.append(impsCh1[i])
    impsCh3.append(sRGB_TRC_encode(impsCh2[2], L_max))
    for i in range(n):
        impsCh3[i].setTitle(chName[i])
        impsCh3[i].show() # < showing imp is need to merge channels
    imp3 = IJ.run("Merge Channels...", "c1=R c2=G c3=B create keep")
    IJ.saveAsTiff(imp3, folder + filename +"_B_level_adjusted.tif")
    for i in range(n):
        impsCh1[i].close()
        impsCh2[i].close()
        impsCh3[i].close()

    IJ.showMessage("B level adjustment Finished")
    imp0.close()
    imp1.close()
else:
    try:
        imp0.close()
    except:
        pass
#---

